# -*- coding: utf-8 -*-

"""pr0xsurv - A package for proximal survival analysis"""

__author__ = """Rahul Goswami"""
__email__ = 'yuvrajiro@gmail.com'
__version__ = '0.0.3'

print(" We are to sorry to inform you we have not yet implemented the package.")